//v1.0.4
//command line:
//node dslsubentries.js input.dsl output.dsl
//input.dsl - файл в формате DSL, UTF8 с BOM или без и UTF16LE только с BOM

'use strict';
const fs = require('fs');
const path = require('path');
const readline = require('readline');
const getByteLength = Buffer.byteLength;

let fileSize = 0;
let lineCount = 0;
let errorCount = 0;



function fileExists(filePath)
{
	try
	{
		return fs.statSync(filePath).isFile();
	}
	catch (err)
	{
		return false;
	}
}


// progress bar as a function (CREDITS: vsemozhetbyt)
//https://gist.github.com/vsemozhetbyt/3b225c62f226e8d71156c6029ce298eb

function pb(edge = 0) {
	const rl = require('readline');

	const DEFAULT_FREQ = 500;
	const HUNDRED_PERCENT = 100;
	const PB_LENGTH = 50;
	const PB_SCALE = HUNDRED_PERCENT / PB_LENGTH;

	const NANOSECONDS_PER_SECOND = 1e9;

	const hrStart = process.hrtime();

	function clearLine() {
		rl.cursorTo(process.stdout, 0);
		rl.clearLine(process.stdout, 0);
	}

	function getTimePast() {
		const hrEnd = process.hrtime(hrStart);
		return `${
			((hrEnd[0] * NANOSECONDS_PER_SECOND + hrEnd[1]) / NANOSECONDS_PER_SECOND).toFixed(1)
		} s`;
	}

	return {
		edge,
			stat: 0,

			start(freq = DEFAULT_FREQ) {
				this.updater = setInterval(() => { this.update(); }, freq);
			},

			update(stat = this.stat) {
				const statPercent =
					stat === this.edge || stat > this.edge ?
					HUNDRED_PERCENT :
					stat / this.edge * HUNDRED_PERCENT;

				const barsNumber = Math.floor(statPercent / PB_SCALE);
				const padsNumber = PB_LENGTH - barsNumber;

				clearLine();
				process.stdout.write(
						`${'█'.repeat(barsNumber)}${'░'.repeat(padsNumber)} ${statPercent.toFixed(1)}%  ${getTimePast()} (${stat.toLocaleString()} of ${this.edge.toLocaleString()})`
						);
			},

			end() {
				clearInterval(this.updater);
				this.stat = this.edge;
				this.update();
				console.log('\n');
			},

			clear() {
				clearInterval(this.updater);
				clearLine();
			},
	};
}



function guessEncoding(path) {
	const BOM_0 = 0xFF;
	const BOM_1 = 0xFE;

	try  {
		const fd = fs.openSync(path, 'r');
		const bf = Buffer.alloc(2);
		fs.readSync(fd, bf, 0, 2, 0);
		fs.closeSync(fd);
		return bf[0] === BOM_0 && bf[1] === BOM_1 ? 'utf16le' : 'utf8';
	} catch (e) {
		console.error(`Error: ${e.message}.`);
		return null;
	}
}


function remove_comments(s)
{
	s = s.replace(/(\\*)(\{\{[^]*?\}\})/g,  function(a, m1, m2){if ((m1.length % 2) === 0)m2 = ''; return m1 + m2; })
		return s;
}



function remove_scb(s)
{

	s = s.replace(/\x00/g, " ").replace(/(\\*)(\{|\})/g,  function(a, m1, m2)
			{


			if (((m1.length % 2) === 0) && (m2 === '{'))
			{

			m2 = "\x00" + m2;
			}
			else if (((m1.length % 2) === 0) && (m2 === '}'))
			{
			m2 = "\x00" + m2;
			}

			return m1 + m2; 

			});



	s = s.replace(/(\x00\{[^\x00]*?\x00\})/g,  function(a, m1){return '';}).replace(/\x00/g, "");

	return s;

}



if ((process.argv.length === 4) && (fileExists(process.argv[2])))
{
	by_dls_article();
}
else
{

	console.log('Invalid command line.');
	process.exit();

}




function NotEscapeDogCount(s)
{

	let m;
	let res = [0, 0];
	let re = /[^]*?(\\*)@/y;
	let ind = 0;
	while (m = re.exec(s))
	{
		if ((m[1].length % 2) === 0)
		{
			res[0]++;
			ind = re.lastIndex;
		}

	}

	res[1] = ind;

	return res;

}



function process_subentry(body, art_start)
{


	let newarts = "";
	let subs = [];

	//[0] - это подкарточка или разделитель; подкарточка - true, разделитель - false 
	//[1] - массив с заголовками
	//[2] - массив со строками тела подкарточки
	//[3] - закрыта ли подкарточка

	for (let i = 0; i < body.length; i++)
	{
		let dogs = NotEscapeDogCount(body[i]);



		if (dogs[0] === 1)
		{


			if (remove_comments(body[i].substr(dogs[1])).trim() === '')
			{

				if ((subs.length > 0) && (subs[subs.length -1][0]) && (subs[subs.length -1][3] === 0) && (subs[subs.length -1][2].length === 0))
				{



					fs.writeFileSync('error.log', `Error: Пустая подкарточка. Line: ${art_start}\n`, {encoding: 'utf8', flag: "a"});
					errorCount++;
					return [false];

				}
				else if ((subs.length > 0) && (subs[subs.length -1][0]) && (subs[subs.length -1][3] === 0) && (subs[subs.length -1][2].length > 0))
				{

					subs[subs.length -1][3] = 1;
					subs.push([false]);

				}
				else if ((subs.length === 0)||((subs.length > 0) && (!subs[subs.length -1][0])))
				{
					fs.writeFileSync('error.log', `Error: Пустой подзаголовок. Line: ${art_start}\n`, {encoding: 'utf8', flag: "a"});
					errorCount++;
					return [false];
				}
				else
				{

					fs.writeFileSync('error.log', `Error. Line: ${art_start}\n`, {encoding: 'utf8', flag: "a"});
					errorCount++;
					return [false];

				}



			}
			else
			{


				let h0 = body[i].substr(dogs[1]);
				let pre = "\n" + body[i].substr(0,dogs[1]-1);
				let h1 = h0;

				let hw2 = [];
				h1 = remove_comments(h1);
				h1 = remove_scb(h1).replace(/[\t ]{2,}/g, " ");
				h0 = h0.replace(/[\t ]{2,}/g, " ");

				if (h1.trim().length > 0)
				{
					newarts += pre + "{{- }}<<" + h1.trim() + ">>";
					hw2.push(h0.trim());
				}

				if (hw2.length === 0)
				{

					fs.writeFileSync('error.log', `Error: Пустой подзаголовок. Line: ${art_start}\n`, {encoding: 'utf8', flag: "a"});
					errorCount++;
					return [false];
				}
				else
				{


					//[0] - это подкарточка или разделитель; подкарточка - true, разделитель - false 
					//[1] - массив с заголовками
					//[2] - массив со строками тела подкарточки
					//[3] - закрыта ли подкарточка



					if ((subs.length > 0) && (subs[subs.length - 1][0]) && (subs[subs.length - 1][1].length > 0) && (subs[subs.length - 1][2].length === 0) && (subs[subs.length - 1][3] === 0))
					{

						subs[subs.length - 1][1].push(...hw2);
					}
					else
					{

						if ((subs.length > 0) && (subs[subs.length - 1][0]) && (subs[subs.length - 1][3] === 0))
							subs[subs.length - 1][3] = 1;

						subs.push([true, hw2, [], 0, i]);
					}


				}

			}


		}
		else if (dogs[0] === 0)
		{
			if ((subs.length > 0) && (subs[subs.length -1][0]) && (subs[subs.length -1][3] === 0) && (body[i].trim().length > 0))
			{
				subs[subs.length -1][2].push(body[i]);
			}
			else
			{
				newarts += "\n" + body[i];
			}
		}
		else
		{
			fs.writeFileSync('error.log', `Error: Неправильное использование символа начала подкарточки "@". Line: ${art_start}\n`, {encoding: 'utf8', flag: "a"});
			errorCount++;
			return [false];
		}


	}


	let hw3 = [];

	
	let foot = newarts.match(/([\s]+)$/) || [undefined, ""];


	for (let v of subs)
	{

		//[0] - это подкарточка или разделитель; подкарточка - true, разделитель - false 
		//[1] - массив с заголовками
		//[2] - массив со строками тела подкарточки
		//[3] - закрыта ли подкарточка


		if ((v[0]) && (v[1].length > 0) && (v[2].length > 0) && (v[3] === 1))
		{


			hw3.push(v[1].join("\n") + "\n" + v[2].join("\n") + foot[1]);

		}
		else
		{

			if (v[0])
			{
				if ((v[3] === 1) && (v[1].length === 0))
				{
					fs.writeFileSync('error.log', `Error: Пустой подзаголовок. Line: ${art_start}\n`, {encoding: 'utf8', flag: "a"});
					errorCount++;
					return [false];
				}
				else if ((v[3] === 1) && (v[2].length === 0))
				{

					fs.writeFileSync('error.log', `Error: Пустая подкарточка. Line: ${art_start}\n`, {encoding: 'utf8', flag: "a"});
					errorCount++;
					return [false];
				}					
				else if ((v[1].length > 0) && (v[3] === 0))
				{
					fs.writeFileSync('error.log', `Error: Подкарточка не окончена. Line: ${art_start}\n`, {encoding: 'utf8', flag: "a"});
					errorCount++;
					return [false];
				}
			}

		}


	}



	if (hw3.length > 0)
		newarts += "\n" + hw3.join("\n");


	return [true, newarts];

}




function ProcessDSLArticle(art0, art1, art_num, art_start)
{

	let hw1 = [];
	let info = [0,0,0,0];
	//[0] - счётчик заголовков, [1] - счётчик непустых строк в теле статьи
	//[3] - счётчик @ в теле

	for (let i = 0; i < art1.length; i++)
	{


		if (/^[^\t\x20]/.test(art1[i]))
		{

			info[0]++;
			hw1.push(art1[i].replace(/[\t ]{2, }/g, " ").trim());

		}
		else if (/^[\t\x20]+[^\t\x20\n]/.test(art1[i]))
		{
			if (info[1] === 0)
				info[2] = i;

			info[1]++

				if (/@/.test(art1[i]))
					info[3]++;

		}


	}



	if (info[0] === 0)
	{
		fs.writeFileSync('error.log', `Error: Пустой заголовок карточки. Line: ${art_start}\n`, {encoding: 'utf8', flag: "a"});
		errorCount++;
	}


	if (info[1] === 0)
	{

		fs.writeFileSync('error.log', `Error: Пустая статья. Line: ${art_start}\n`, {encoding: 'utf8', flag: "a"});
		errorCount++;

	}


	let artstr = art0.join("\n");


	if ((info[0] > 0) && (info[1] > 0) && info[3] > 0)
	{


		let hw0 = art0.slice(0,info[2]);
		let body = art0.slice(info[2]);


		let r = process_subentry(body, art_start);


		if (r[0])
		{
			return  hw0.join("\n") + "\n" + r[1].replace(/^[\r\n]*/, '');
		}
		else
		{
			return artstr;
		}


	}

	return  artstr;      	

}


function parseLine(line)
{

	let m;
	let re = /([^]*?)(\\*)(\{\{|\}\})/y;
	let e;
	let l;

	while (m = re.exec(line))
	{

		if ((m[2].length % 2) === 1)
		{
			continue;
		}

		if (m[3] === '}}')
{

	l = [false, re.lastIndex]
		if (!e)
		{
			e = [true, re.lastIndex];
		}

}
else
{
	l = [true, re.lastIndex]
		if (!e)
		{
			e = [false, re.lastIndex];
		}


}

}


let r = [[false, undefined, undefined], [false, undefined]];

if (e)
{
	r[0][0] = true;
	r[0][1] = e[0];
	r[0][2] = e[1];
}

if ((l) && (l[0]))
{
	r[1][0] = l[0];
	r[1][1] = l[1];

}


//Описание ответа:
//r[0][0] - информация о наличие закрытие тега {{
//r[0][1] - если true, то тег закрылся нормально. Нет - значит через {{
//r[0][2] - позиция закрывающего относительно начала строки

//r[1][0] - закончилась ли строка открывающим тегом
//r[1][1] - если да, то передаётся и позиция этого тега

return r;

}

function by_dls_article()
{

	let commLines = [];

	let lineCount = 0;

	let art_start = 0;

	let articleCount = 0;

	let inputFile = process.argv[2];
	let outputFile = process.argv[3];


	let encoding = guessEncoding(inputFile);

	fileSize  = fs.statSync(inputFile)['size'];

	const reader = readline.createInterface({
input: fs.createReadStream(inputFile, encoding),
terminal: false,
historySize: 0
});

const output = fs.openSync(outputFile, 'w');

let flag = 0;


let art0 = [];
let art1 = [];
let lines = [];


fs.writeSync(output, '\uFEFF', null,  encoding);

fs.writeFileSync('error.log', "", {encoding: 'utf8', flag: "w"});

const pbAsync = pb(fs.statSync(inputFile).size);

pbAsync.start();

reader.on('line', line => {


		lineCount++;

		pbAsync.stat += getByteLength(line, encoding) + 1;


		if (lineCount === 1) line = line.replace(/^\uFEFF/, '');

		let r = parseLine(line);


		//Описание ответа:
		//r[0][0] - информация о наличие закрытие тега {{
		//r[0][1] - если true, то тег закрылся нормально. Нет - значит через {{
		//r[0][2] - позиция закрывающего относительно начала строки

		//r[1][0] - закончилась ли строка открывающим тегом
		//r[1][1] - если да, то передаётся и позиция этого тега


		if (commLines.length > 0)
		{

			if ((r[0][0]) && (!r[0][1]) && (r[1][0]))
			{
				//{{ ... {{

				lines.push(...commLines);
				commLines.length = 0;
				commLines.push([lineCount, line]);

			}
			else if ((r[0][0]) && (!r[0][1]) && (!r[1][0]))
			{
				//{{ ... 

				lines.push(...commLines);
				commLines.length = 0;
				lines.push([lineCount, line]);

			}
			else if ((r[0][0]) && (r[0][1]) && (!r[1][0]))
			{
				//}} ...

			commLines.push([lineCount, line]);
			lines.push(commLines.shift());
			while (commLines.length > 0)
			{
				lines[lines.length-1][1] += "\n" + commLines.shift()[1];
			}

			}
			else if ((r[0][0]) && (r[0][1]) && (r[1][0]))
			{
				//}} ... {{

			while (commLines.length > 1)
			{
				let last = commLines.pop();
				commLines[commLines.length - 1][1] += "\n" + last[1];

			}

			commLines.push([lineCount, line]);

			}
			else
			{

				commLines.push([lineCount, line]);

			}



			}
			else
			{


				if (r[1][0])
				{
					commLines.push([lineCount, line]);
				}
				else
				{
					lines.push([lineCount, line]);
				}


			}



			while (lines.length)
			{
				let s = [];
				let l = lines.shift();
				let n = l[0];
				s[0] = l[1];
				s[1] = remove_comments(s[0]);



				if ((/^#/.test(s[1])) && (flag === 0))
				{
					fs.writeSync(output, `${s[0]}\n`, null, encoding);

				}
				else if ((/^\s*$/.test(s[1])) && (flag === 0))
				{
					fs.writeSync(output, `${s[0]}\n`, null, encoding);
				}
				else if ((s[1] === '') &&  (flag !== 0))
				{
					art0.push(s[0]);
					art1.push(s[1]);
					if (flag === 2)
						flag++;

				}
				else if (((/^[^\t ].*$/.test(s[1])) && (flag !== 1)) || ((/^[\t ].*[^\t ].*$/.test(s[1]))&&((art1.length === 0)||(flag === 3))))
				{


					if (art0.length > 0)
					{


						articleCount++;

						const res = ProcessDSLArticle(art0, art1, articleCount, art_start);


						if (res !== null)
						{
							fs.writeSync(output, res + "\n", null, encoding);

						}


					}

					art_start = n;

					art0.length = 0;
					art1.length = 0;


					art0.push(s[0]);
					art1.push(s[1]);


					if (/^[^\t ].*$/.test(s[1]))
					{
						flag = 1;
					}
					else
					{
						flag = 2;
					} 


				}
				else
				{

					art0.push(s[0]);
					art1.push(s[1]);

					if (/^[^\t ].*$/.test(s[1]))
					{
						flag = 1;
					}
					else
					{
						flag = 2;
					} 


				}

			}


			}).on('close', () => {


				lines.push(...commLines);
				commLines.length = 0;

				while (lines.length)
				{

				let s = [];
				let l = lines.shift();
				let n = l[0];
				s[0] = l[1];
				s[1] = remove_comments(s[0]);


				if ((/^#/.test(s[1])) && (flag === 0))
				{
				fs.writeSync(output, `${s[0]}\n`, null, encoding);

				}
				else if ((/^\s*$/.test(s[1])) && (flag === 0))
				{
					fs.writeSync(output, `${s[0]}\n`, null, encoding);
				}
				else if ((s[1] === '') && (flag !== 0))
				{
					art0.push(s[0]);
					art1.push(s[1]);
					if (flag === 2)
						flag++;


				}
				else if (((/^[^\t ].*$/.test(s[1])) && (flag !== 1)) || ((/^[\t ].*[^\t ].*$/.test(s[1]))&&((art1.length === 0)||(flag === 3))))
				{

					if (art0.length > 0)
					{


						articleCount++;

						const res = ProcessDSLArticle(art0, art1, articleCount, art_start);

						if (res !== null)
						{
							fs.writeSync(output, res + "\n", null, encoding);

						}


					}

					art_start = n;

					art0.length = 0;
					art1.length = 0;


					art0.push(s[0]);
					art1.push(s[1]);


					if (/^[^\t ].*$/.test(s[1]))
					{
						flag = 1;
					}
					else
					{
						flag = 2;
					} 


				}
				else
				{
					art0.push(s[0]);
					art1.push(s[1]);

					if (/^[^\t ].*$/.test(s[1]))
					{
						flag = 1;
					}
					else
					{
						flag = 2;
					} 


				}

				}


				if (art0.length > 0)
				{

					articleCount++;

					const res = ProcessDSLArticle(art0, art1, articleCount, art_start);

					if (res !== null)
					{
						fs.writeSync(output, res + "\n", null, encoding);
					}

					art0.length = 0;
					art1.length = 0;

				}


				pbAsync.end();

				if (errorCount > 0)
					console.log("Number of Errors: " + errorCount + " (Check error.log)");


			});



			}

